# core/plugs.py
#
#

""" Plugins package, loads all plugins available. """

__copyright__ = "Copyright 2014 B.H.J Thate"

## IMPORTS

from core.utils import error
from core import Object

import importlib
import logging
import pkgutil

## PLUGINS

class Plugins(Object):

    """ object to register plugins with. """

    def __init__(zelf, *args, **kwargs):
        Object.__init__(zelf, *args, **kwargs)
        zelf.plugins = []

    def load_all(zelf, modname):
        """ load plugins from the plugin directory. """
        logging.warn("loading %s" % modname)
        mod = zelf.load(modname)
        for pkg in pkgutil.walk_packages(mod.__path__, mod.__name__+'.'):
            mod = zelf.load(pkg[1])
            zelf.plugins.append(mod.__name__.split(".")[-1])
        zelf.ready()

    def load(zelf, plugname, force=True):
        """ load module. """
        if plugname in zelf and not force: return zelf.reload(plugname)
        zelf[plugname] = importlib.import_module(plugname)
        return zelf[plugname]

    def reload(zelf, plugname):
        zelf[plugname] = importlib.reload(plugname)
        return zelf[plugname]
 
    def init(zelf, plugname):
        """ initialize plugin. """
        try: zelf[plugname].init()
        except (KeyError, AttributeError): pass
            
    def shutdown(zelf, plugname): 
        """ call the shutdown() function, if provided. """
        try: zelf[plugname].shutdown()
        except (KeyError, AttributeError): pass
        
    def exit(zelf, *args, **kwargs):
        """ exit all plugins. """
        for plugname in zelf.plugins: zelf.shutdown(plugname)
